###############################################################################
#
#  CppOptimizationMethod.py --
#
#  $LastChangedDate$
#  $LastChangedRevision$
#  $LastChangedBy$
#
#  Author: ltomaso
#
#  History: 
#    March 5, 2013 - initial revision
#    May 13, 2013 - switch to the new API (Ansys.DesignXplorer.API.Common).
#                 - IDXOptimizationServices has been renamed IOptimizationServices and derives from IServices
#                 - IDXOptimizationMethod has been renamed IOptimizationMethod and derives from IMethod
#                 - IDXOptimizationPoint has been renamed IOptimizationPoint and derives from IPoint
#
###############################################################################

clr.AddReference('Ans.ComponentSystem.Interop')
from Ans.ComponentSystem import (
    IListCpp, VariantColl, IEnumeratorCpp
)

import os
import string

clr.AddReference('Ans.Utilities')
dxpath = Ansys.Utilities.ApplicationConfiguration.DefaultConfiguration.GetModuleBinDir('Addins/DesignXplorer')
sys.path.append(dxpath)

clr.AddReference('Ans.DesignXplorer.PublicAPIs')
from Ansys.DesignXplorer.API.Common import (
    IServices, IMethod,
    IPoint, enumPointState
)

from Ansys.DesignXplorer.API.Optimization import (
    IOptimizationServices, IOptimizationMethod,
    IOptimizationPoint, DXOptimizationPoint,
	IConvergenceData, DXConvergenceData,
	IConvergenceDescription, DXConvergenceDescription,
    enumGoalType, enumConstraintType,
    enumPostProcessingType,
	enumCurveType
)

import random
import math

import ctypes

# Parameter class for continuous input parameter (double variable)
# continuous input parameters with manufacturable values (double list variable) are not supported by this optimizer
# discrete input parameters (integer list variable) are not supported by this optimizer
class DoubleVariable(object):
    def __init__(self,id,lb,ub,iv):
         self._id = id
         self._lowerbound = lb
         self._upperbound = ub
         # current value of the starting point (if the optimizer starts from an initial point)
         self._initialvalue = iv  

# Parameter class for continuous input parameter (double variable)
# continuous input parameters with manufacturable values (double list variable) are not supported by this optimizer
# discrete input parameters (integer list variable) are not supported by this optimizer
class DoubleListVariable(object):
    def __init__(self,id,vals,iv):
         self._id = id

         numvals = IListCpp.Count(vals);
         self._values = (ctypes.c_double * nbvals)()
         enume = IListCpp.GetEnumerator(vals)
         i = 0
         while IEnumeratorCpp.MoveNext(enume):
            val = IEnumeratorCpp.Current.GetValue(enume)
            self._values[i] = ctypes.c_double(val);
            i = i + 1
         # current value of the starting point (if the optimizer starts from an initial point)
         self._initialvalue = ctypes.c_double(iv);  
         
      

# Objective class
class Objective(object):
     def __init__(self,id,type,val,imp):
         self._id = id
         self._goalvalue = val
         self._importance = imp
         self._type = 0
         if type == enumGoalType.eGT_MinimumPossible:
             self._type = 1
         elif type == enumGoalType.eGT_MaximumPossible:
             self._type = 2
         elif type == enumGoalType.eGT_SeekTarget:
             self._type = 3

# Constraint class
class Constraint(object):
      def __init__(self,id,type,val1,val2,imp,strict):
         self._id = id
         self._type = 0
         if type == enumConstraintType.eCT_LessThanTarget:
              self._type = 1
         elif type == enumConstraintType.eCT_GreaterThanTarget:
              self._type = 2
         elif type == enumConstraintType.eCT_NearTarget:
              self._type = 3
         elif type == enumConstraintType.eCT_InsideBounds:
             self._type = 4

         self._value1 = val1
         self._value2 = val2
         self._importance = imp
         #self._strict = strict
         
# PythonOptimizer class
class PythonWrapperOptimizer(IOptimizationMethod):
    # initializes PythonOptimizer
    def __init__(self):
        global opt

        self._outputs = enumPostProcessingType.ePPT_Candidates | enumPostProcessingType.ePPT_Samples
        self._settings = {}
        self._candidates = None
        self._samples = VariantColl()

        self._parameters = []
        #self._outputs = []
        self._inputparameters = []
        self._outputparameters = []
        self._doublevars = []
        self._doublelistvars = []
        self._integerlistvars = []
        self._objectives = []
        self._constraints = []
        self._services = None

        self._init = False;
        
    # sets IOptimizationServices
    def set_Services(self,value):
        self._services = value
        

    # returns value of optimization setting
    def get_Setting(self,key):
        return self._settings[key]
        
    # puts value of optimization setting
    def put_Setting(self,key,value):
        self._settings[key] = value
     
    # returns post-processing type   
    def get_PostProcessingTypes(self):
        return self._outputs

    # returns candidate points
    def get_Candidates(self):
        return self._candidates     
    
    #returns sample points    
    def get_Samples(self):
        return self._samples
        
    # definition of double variable
    def AddDoubleVariable(self,id,lb,ub,iv):
        p = DoubleVariable(id,lb,ub,iv)
        self._doublevars.append(p)
        self._inputparameters.append(id)
        self._parameters.append(id)
    
    # integer list variable is ignored: not supported by this optimizer                   
    def AddIntegerListVariable(self,id,vals,iv):
        #p = ListVariable(id,vals,iv)
        #self._integerlistvars.append(p)
        #self._inputparameters.append(id)
        self._parameters.append(id)
        pass
      
    # double list variable is ignored: not supported by this optimizer  
    def AddDoubleListVariable(self,id,vals,iv):
        #p = DoubleListVariable(id,vals,iv)
        #self._doublelistvars.append(p)
        #self._inputparameters.append(id)
        self._parameters.append(id)
        
    # definition of output parameter is ignored
    def AddOutput(self,id):
        self._outputparameters.append(id) 
        self._parameters.append(id)                  
    
    # creates an objective and adds it to the list of objectives
    def AddObjective(self,id,type,val,imp):
        o = Objective(id,type,val,imp)
        self._objectives.append(o)
    
    # creates a constraint and adds it to the list of constraints
    def AddConstraint(self,id,type,val1,val2,imp,strict):
        c = Constraint(id,type,val1,val2,imp,strict)
        self._constraints.append(c)
    
    # returns true when all optimization settings are defined and correct and the optimizer is ready to run        
    def CanRun(self, errorMessage):
        return True

    # runs optimization 
    def Run(self):
        ExtAPI.Log.WriteMessage('Run')
        #System.Diagnostics.Debugger.Break()

        # initializes the C++ optimizer
        self.InitOptimizer()
        

        # resets the list of candidate points 
        self._candidates = None

        global opt

        numinputs = len(self._inputparameters)
        # runs C++ optimizer
        iter = 1
        numpts = opt.RunOptimization()
        while numpts > 0:
            ipt = 0
            pts = []
            # gets input parameters values from C++ optimizer
            while ipt < numpts:
                dblvals = (ctypes.c_double * numinputs)()
                opt.GetInputValues(ipt,ctypes.cast(dblvals,ctypes.POINTER(ctypes.c_double)))
                pts.append(dblvals)
                ipt = ipt + 1

            # converts list of double values to list of optimization points
            pointsToCalculate = self.GenerateOptimizationPoints(pts,True)

            # calculates optimization points
            IOptimizationServices.CalculatePoints(self._services,pointsToCalculate)

            # pushes calculated optimization points to C++ optimizer
            self.PushOutputValues(pointsToCalculate)

            # promotes all points as history points
            enume = pointsToCalculate.GetEnumerator()
            while IEnumeratorCpp.MoveNext(enume):
                optpt = IEnumeratorCpp.Current.GetValue(enume)
                if optpt.State == enumPointState.ePS_UpToDate:
                    IOptimizationServices.PushHistoryPoint(self._services,optpt)
                    self._samples.Add(optpt)

            if opt.GetNumberOfGeneratedCanidatePoints() > 0:
                feasible = (ctypes.c_bool * 1)()
                cost = (ctypes.c_double * 1)()
                opt.GetCostFunctionValues(ctypes.cast(cost,ctypes.POINTER(ctypes.c_double)),ctypes.cast(feasible,ctypes.POINTER(ctypes.c_bool)))
                data = DXConvergenceData()
                data.put_Value(0,iter,cost[0],feasible[0])
                IOptimizationServices.PushConvergenceData(self._services,data)
            # stops the optimization algorithm if the user has stopped the optimization run
            if IOptimizationServices.Stopped.GetValue(self._services):
                break

            # runs C++ optimizer
            numpts = opt.RunOptimization()
            iter = iter + 1

        # end of C++ optimization
        # gets candidate points from C++ optimizer
        numcandidates = opt.GetNumberOfGeneratedCanidatePoints()
        ic = 0
        numparams = len(self._parameters)
        candidates = []
        while ic < numcandidates:
            dblvals = (ctypes.c_double * numparams)()
            opt.GetCandidatePointValues(ic,ctypes.cast(dblvals,ctypes.POINTER(ctypes.c_double)))
            candidates.append(dblvals)
            ic = ic + 1

        # converts list of candidate values to list of optimization points
        self._candidates = self.GenerateOptimizationPoints(candidates,False)

        # output settings    
        self._settings["MyOptimizationStatus"] = opt.GetStatus()

    def InitOptimizer(self):
         global opt
         # checks whether the C++ optimizer has already been created and initialized
         if self._init == True:
             pass
         
         # creates optimizer

         # definition of variables
         doubleVarIdx = 0
         doubleListVarIdx = 0
         integerListVarIdx = 0

         ip = 0
         # loop on input parameters
         for input in self._inputparameters:
             # double variable ?
             if doubleVarIdx < self._doublevars.Count and self._doublevars[doubleVarIdx]._id == input:
                 # adds double variable
                 var = self._doublevars[doubleVarIdx]
                 opt.AddDoubleVariable(ip,var._lowerbound,var._upperbound,var._initialvalue)
                 doubleVarIdx = doubleVarIdx + 1
             else:
                 # double list variable ?
                 if doubleListVarIdx < self._doublelistvars.Count and self._doublelistvars[doubleListVarIdx]._id == input:
                     # adds double list variable
                     var = self._doublelistvars[doubleListVarIdx]
                     opt.AddDoubleListVariable(ip,ctypes.cast(var._values,ctypes.POINTER(ctypes.c_double)), len(var._values),var._initialvalue)
                     doubleVarIdx = doubleVarIdx + 1
                 #else:
                 #    # integer list variable ?
                 #    if integerListVarIdx < self._integerlistvars.Count and self._integerlistvars[integerListVarIdx]._id == input:
                 #    #add integer list variable
                 #    var = self._integerlistvars[integerListVarIdx]
                 #    self._opt.AddDoubleVariable(input,ctypes.cast(var._values,ctypes.POINTER(ctypes.c_double)), len(var._values),var._initialvalue)
                 #    integerListVarIdx = integerListVarIdx + 1
             ip = ip + 1

         #opt.PutNumberOfOutputs(len(_outputparameters))

         ip = 0
         numinputs = len(self._inputparameters)
         numoutputs = len(self._outputparameters)
         iobj = 0
         iconstr = 0
         for param in self._parameters:
             if iobj < self._objectives.Count and param == self._objectives[iobj]._id:
                 opt.AddObjective(ip,self._objectives[iobj]._type,self._objectives[iobj]._goalvalue,self._objectives[iobj]._importance)
                 iobj = iobj + 1
             if iconstr < self._constraints.Count and param == self._constraints[iconstr]._id:
                 opt.AddConstraint(ip,self._constraints[iconstr]._type,self._constraints[iconstr]._value1, self._constraints[iconstr]._value2, self._constraints[iconstr]._importance)#, self._constraints[iconstr]._strict)
                 iconstr = iconstr + 1
             ip = ip + 1

         # forwards optimization settings to the C++ optimizer
         self.PutSettings()
         self._init = True
    
    # forwards optimization settings to the C++ optimizer
    def PutSettings(self):
        global opt
        for key in self._settings:
            if key == 'MyNumberOfSamples':
                opt.PutNumberOfSamples(self._settings[key])
            if key == 'MyMaximumNumberOfCandidates':
                opt.PutMaximumNumberOfCandidates(self._settings[key])

    # converts list of double values to list of optimization points (with or without output parameter values)
    def GenerateOptimizationPoints(self,values,onlyinputs):
        numpts = len(values)
        if numpts == 0:
            return null

        points = VariantColl()
        ptIdx = 0
        for ptvalues in values:
            point = DXOptimizationPoint()
            point.ID = 'DP %(id)04d' % {"id" : ptIdx}
            ip = 0
            for input in self._inputparameters:
                var = ptvalues[ip]
                point.put_Value(input,var)
                ip = ip + 1
            if onlyinputs == False:
                for output in self._outputparameters:
                    var = ptvalues[ip]
                    IOptimizationPoint.put_Value(point, output,var)
                    ip = ip + 1
            points.Add(point)
        return points

    # converts calculated optimization points to list of double values to push them into the C++ optimizer
    def PushOutputValues(self,points):
        global opt
        numoutputs = len(self._outputparameters)
        ip = 0
        enume = points.GetEnumerator()
        while IEnumeratorCpp.MoveNext(enume):
            pt = IEnumeratorCpp.Current.GetValue(enume)
            iout = 0
            dblvals = (ctypes.c_double * numoutputs)()
            state = False
            if pt.State == enumPointState.ePS_UpToDate:
                state = True
                for output in self._outputparameters:
                    var = IOptimizationPoint.get_Value(pt, output)
                    dblvals[iout] = ctypes.c_double(var)
                    iout = iout + 1
            opt.AddOutputValues(ip, state, ctypes.cast(dblvals,ctypes.POINTER(ctypes.c_double)),numoutputs)
            ip = ip + 1
         
# end of the class

