###############################################################################
#
#  OptimizerMethod.py --
#
#  $LastChangedDate$
#  $LastChangedRevision$
#  $LastChangedBy$
#
#  Author: jmblanch
#
#  History: 
#    March 5, 2013 - initial revision
#
###############################################################################

clr.AddReference('Ans.ComponentSystem.Interop')
from Ans.ComponentSystem import (
    IListCpp, VariantColl, IEnumeratorCpp
)

import os
import string

clr.AddReference('Ans.Utilities')
dxpath = Ansys.Utilities.ApplicationConfiguration.DefaultConfiguration.GetModuleBinDir('Addins/DesignXplorer')
sys.path.append(dxpath)
ExtAPI.Log.WriteMessage(dxpath)

clr.AddReference('Ans.DesignXplorer.PublicAPIs')

from Ansys.DesignXplorer.API import (
    ParameterNature
)

from Ansys.DesignXplorer.API.Common import (
    IServices, IMethod,
    IPoint, IUserExtensionObject,
    IUserExtensionOptimizationCriterion, IUserExtensionParameter,
    enumPointState, enumSimulationType, enumDistributionType
)

from Ansys.DesignXplorer.API.Optimization import (
    IOptimizationServices, IOptimizationMethod,
    IOptimizationPoint, DXOptimizationPoint,
	IConvergenceData, DXConvergenceData,
	IConvergenceDescription, DXConvergenceDescription,
    enumGoalType, enumConstraintType,
    enumPostProcessingType,
	enumCurveType
)

import random
import math

import OptimizationHelper


# Parameter class for continuous input parameter (double variable)
# continuous input parameters with manufacturable values (double list variable) are not supported by this optimizer
# discrete input parameters (integer list variable) are not supported by this optimizer
class Parameter(object):
    def __init__(self,id,type,lb,ub,iv):
         self._id = id
         self._type = type
         self._lowerbound = lb
         self._upperbound = ub
         # current value of the starting point (if the optimizer starts from an initial point)
         self._initialvalue = iv   
         self._customproperties = {}
         
# Objective class
class Objective(object):
     def __init__(self,id,type,val,imp):
         self._id = id
         self._type = type
         self._goalvalue = val
         self._importance = imp
         self._customproperties = {}

     # cost function of the objective
     def Compute(self,pt):
         id = self._id
         cost = 0.0
         if self._type == enumGoalType.eGT_MinimumPossible:  
             cost = pt.get_Value(id) * self._importance
         elif self._type == enumGoalType.eGT_MaximumPossible:
             cost = -pt.get_Value(id) * self._importance
         elif self._type == enumGoalType.eGT_SeekTarget:
             diff = math.fabs(self._goalvalue - pt.get_Value(id))
             if self._customproperties != {} and self._customproperties["AbsoluteCutOffTolerance"] != "":
                tol = self._customproperties["AbsoluteCutOffTolerance"]
                if diff > tol:
                    diff = diff + 1000 # add penalty term
             cost = diff * self._importance
         return cost

# Constraint class
class Constraint(object):
      def __init__(self,id,type,val1,val2,imp,strict):
         self._id = id
         self._type = type
         self._value1 = val1
         self._value2 = val2
         self._importance = imp
         self._strict = strict
         self._customproperties = {}
         
      # cost function of the constraint
      def Compute(self,pt):
          id = self._id
          violation = 0.0
          if self._type == enumConstraintType.eCT_LessThanTarget:
              violation = pt.get_Value(id) - self._value1
          elif self._type == enumConstraintType.eCT_GreaterThanTarget:
              violation = self._value1 - pt.get_Value(id)
          elif self._type == enumConstraintType.eCT_NearTarget:
              violation = math.fabs(pt.get_Value(id) - self._value1)
          elif self._type == enumConstraintType.eCT_InsideBounds:
              c1 = self._value1 - pt.get_Value(id)
              c2 = pt.get_Value(id) - self._value2
              if c1 > c2:
                  violation = c1
              else:
                  violation = c2
          violation = violation * self._importance
          feasibleTolerance = 0.0
          if self._customproperties != {} and self._customproperties["FeasibleTolerance"] != "":
            feasibleTolerance = self._customproperties["FeasibleTolerance"]
          
          if violation > feasibleTolerance and self._strict == True:
              violation = violation + 1000 # add penalty term
              
          if violation <= feasibleTolerance:
            violation = 0.0
          return violation

# CostPoint class: stores the cost function value of the optimization point and compares it with any other optimization point
class CostPoint(object):
      def __init__(self,pt,cost):
         self._pt = pt
         self._cost = cost
      def __lt__(self,other):
         return self._cost < other._cost            
      def __gt__(self,other):
         return self._cost > other._cost            
      def __eq__(self,other):
         return self._cost == other._cost            
      def __le__(self,other):
         return self._cost <= other._cost            
      def __ge__(self,other):
         return self._cost >= other._cost            
      def __ne__(self,other):
         return self._cost != other._cost            

# PythonOptimizer class
class PythonOptimizer(IOptimizationMethod):
    # initialization of PythonOptimizer
    def __init__(self, seed, outputs):
        random.seed(seed)
        self._outputs = enumPostProcessingType.ePPT_Candidates
        if outputs == "Candidates":
            self._outputs = enumPostProcessingType.ePPT_Candidates
        elif outputs == "Samples":
             self._outputs = enumPostProcessingType.ePPT_Samples
        elif outputs == "CandidatesAndSamples":
             self._outputs = enumPostProcessingType.ePPT_Candidates | enumPostProcessingType.ePPT_Samples
        elif outputs == "SamplesAndParetos":
             self._outputs = enumPostProcessingType.ePPT_Samples | enumPostProcessingType.ePPT_FeasibilityAndParetoFronts 
        elif outputs == "CandidatesAndSamplesAndParetos":
             self._outputs = enumPostProcessingType.ePPT_Candidates | enumPostProcessingType.ePPT_Samples | enumPostProcessingType.ePPT_FeasibilityAndParetoFronts 
        self._settings = {}
        self._candidates = None
        self._samples = None
        self._parameters = []
        self._objectives = []
        self._constraints = []
        self._services = None
        
    # sets IOptimizationServices
    def set_Services(self,value):
        self._services = ServicesHelper(value)

    # returns value of optimization setting
    def get_Setting(self,key):
        return self._settings[key]
        
    # puts value of optimization setting
    def put_Setting(self,key,value):
        self._settings[key] = value
     
    # returns post-processing type   
    def get_PostProcessingTypes(self):
        return self._outputs

    # returns candidate points
    def get_Candidates(self):
        return self._candidates  

    # definition of AddParameterRelationship		
    def AddParameterRelationship(self, var_id, left_exp, right_exp, par_rel_type):
        pass

    #returns sample points    
    def get_Samples(self):
        return self._samples

    # definition of double variable
    def AddDoubleVariable(self,id,lb,ub,iv):
        p = Parameter(id,2,lb,ub,iv)
        self._parameters.append(p)
    
    # integer list variable is ignored: not supported by this optimizer                   
    def AddIntegerListVariable(self,id,vals,iv):
        pass
      
    # double list variable is ignored: not supported by this optimizer  
    def AddDoubleListVariable(self,id,vals,iv):
        pass
       
    # definition of output parameter is ignored
    def AddOutput(self,id):
        pass
    
    def AddCustomVariableProperty(self,id,name,val):
        for p in self._parameters:
            if p._id == id:
                p._customproperties[name] = val
                break
    
    def AddCustomObjectiveProperty(self,id,name,val):
        for obj in self._objectives:
            if obj._id == id:
                if name == "RelativeObjectiveWeight":
                    obj._importance = val
                else:
                    obj._customproperties[name] = val
                break
    
    def AddCustomConstraintProperty(self,id,name,val):
        for constr in self._constraints:
            if constr._id == id:
                if name == "RelativeConstraintWeight":
                    constr._importance = val
                else:
                    constr._customproperties[name] = val
                break
    
    # creates an objective and adds it to the list of objectives
    def AddObjective(self,id,type,val,imp):
        o = Objective(id,type,val,imp)
        self._objectives.append(o)
    
    # creates a constraint and adds it to the list of constraints
    def AddConstraint(self,id,type,val1,val2,imp,strict):
        c = Constraint(id,type,val1,val2,imp,strict)
        self._constraints.append(c)
    
    # returns true when all optimization settings are defined and correct and the optimizer is ready to run        
    def CanRun(self, errorMessage):
        if len(self._parameters) == 0:
             errorMessage.Value = "You need to specify one parameter at least"
             return False
             
        if len(self._objectives) == 0:
             errorMessage.Value = "You need to specify one objective at least"
             return False
             
        maximumNumberOfCandidates = self._settings["MyMaximumNumberOfCandidates"]
        numberOfSamples = self._settings["MyNumberOfSamples"]
        
        # checks the consistency between the number of sample points and the number of candidate points
        if maximumNumberOfCandidates <= 0 or maximumNumberOfCandidates > numberOfSamples:
             errorMessage.Value = "Maximum Number of Candidates must be positive and lower than or equal to the Number of Samples."
             return False
        
        # numberOfSamples must be positive and less than 10000
        if numberOfSamples <= 0 or numberOfSamples > 10000:
             errorMessage.Value = "Number of Samples must be positive and lower than or equal to 10000."
             return False

        # the optimization can be run
        errorMessage.Value = ""
        return True

    # runs optimization
    def Run(self):
        #System.Diagnostics.Debugger.Break()

        # resets the list of candidate points 
        self._candidates = None
        # log message
        self._services.PublishLogMessage("Generation of random points...\n")
        
        numberOfSamples = self._settings["MyNumberOfSamples"]
        numberOfEvals = 0
        points = VariantColl()
        # generates input values of random points
        for i in range(numberOfSamples):
            pt = self.GenerateRandomPoint(i)
            IListCpp.Add(points,pt)
        
        # calculates optimization points
        self._services.PublishLogMessage("Calculation of points...\n")

        # this optimizer submits a maximum of 10 optimization points
        if 10 > numberOfSamples:
             nbsets = numberOfSamples
        else:
             nbsets = 10
        nbpointsbyset = int(numberOfSamples/nbsets)
        pointsToCalculate = VariantColl()
        ind = 0
        iset = 1
        costpoints = []
        globalmin = 9999
        while ind < numberOfSamples:
             IListCpp.Clear(pointsToCalculate)
             for j in range(nbpointsbyset):
                 if ind >= numberOfSamples:
                    break
                 IListCpp.Add(pointsToCalculate,IListCpp.__getitem__(points,ind))
                 ind = ind + 1

             # calculates the set of optimization points
             self._services.CalculatePoints(pointsToCalculate)

             # publishes progress
             self._services.PublishProgress('Set no%(number)02d calculated' % {"number": iset }, 10)
             data = DXConvergenceData()
             localmin = 9999
             index = 0
             for i in range(IListCpp.Count.GetValue(pointsToCalculate)):
                 pt = IListCpp.__getitem__(pointsToCalculate,index)
                 index = index + 1
                 if pt.State == enumPointState.ePS_UpToDate:
                    numberOfEvals = numberOfEvals + 1
                    # promotes optimization point as history point
                    self._services.PushHistoryPoint(pt)
                    # calculates the cost function value of the optimization point
                    cp = CostPoint(pt,self.CostFunction(pt))
                    if localmin > self.CostFunction(pt):
                       localmin = self.CostFunction(pt)
                    costpoints.append(cp)
             if globalmin > localmin:
                 globalmin = localmin
                 Bool = True
             else:
                 Bool = False
             IConvergenceData.put_Value(data,0,iset,globalmin,Bool)
             IConvergenceData.put_Value(data,1,iset,0.3)
             IConvergenceData.put_Value(data,2,iset,localmin)
             IConvergenceData.put_Value(data,3,iset,localmin+globalmin,Bool)
             IConvergenceData.put_Value(data,4,iset,0)
             IConvergenceData.put_Value(data,5,iset,localmin-globalmin)
             self._services.PushConvergenceData(data)
             iset = iset + 1

             # stops the optimization algorithm if the user has stopped the optimization run
             if self._services.Stopped:
                 break
        # output settings           
        self._settings["MyNumberOfEvaluations"] = numberOfEvals
        if numberOfEvals < numberOfSamples:
            self._settings["MyOptimizationStatus"] = 'Optimization Interrupted after %d evaluations' %(numberOfEvals)
        else:
            self._settings["MyOptimizationStatus"] = 'Optimization Complete'
       
        # using of existing design points (come from data transfer)
        existingPts = self._services.ExistingPoints
        if existingPts != None:
            index = 0
            for i in range(IListCpp.Count.GetValue(existingPts)):
                pt = IListCpp.__getitem__(existingPts,index)
                index = index + 1   
                if pt.State == enumPointState.ePS_UpToDate:
                    # promotes optimization point as history point
                    self._services.PushHistoryPoint(pt)
                    # calculates the cost function value of the optimization point
                    cp = CostPoint(pt,self.CostFunction(pt))
                    if localmin > self.CostFunction(pt):
                       localmin = self.CostFunction(pt)
                    IListCpp.Add(points,pt)
                    costpoints.append(cp)
            if globalmin > localmin:
                globalmin = localmin
                Bool = True
            else:
                Bool = False
            IConvergenceData.put_Value(data,0,iset,globalmin,Bool)
            IConvergenceData.put_Value(data,1,iset,0.3)
            IConvergenceData.put_Value(data,2,iset,localmin)
            IConvergenceData.put_Value(data,3,iset,localmin+globalmin,Bool)
            IConvergenceData.put_Value(data,4,iset,0)
            IConvergenceData.put_Value(data,5,iset,localmin-globalmin)
            self._services.PushConvergenceData(data)
            iset = iset + 1

        # generates candidate points
        if (self._outputs & enumPostProcessingType.ePPT_Candidates) == enumPostProcessingType.ePPT_Candidates:
           self._services.PublishLogMessage("Sorting of points : generation of candidate points...\n")
           costpoints.sort() 
           self._candidates = VariantColl()
           maximumNumberOfCandidates = self._settings["MyMaximumNumberOfCandidates"]
           ind = 0
           for ip in range(maximumNumberOfCandidates):
               IListCpp.Add(self._candidates,costpoints[ind]._pt.Point)
               ind = ind + 1

        # generates sample points
        if (self._outputs & enumPostProcessingType.ePPT_Samples) == enumPostProcessingType.ePPT_Samples:
            Samples = VariantColl()
            ind = 0
            for ip in range(IListCpp.Count.GetValue(points)):
                IListCpp.Add(Samples,IListCpp.__getitem__(points,ind).Point)
                ind = ind + 1
            self._samples = Samples
            if (self._outputs & enumPostProcessingType.ePPT_FeasibilityAndParetoFronts) == enumPostProcessingType.ePPT_FeasibilityAndParetoFronts:
                numFeasibles = numberOfSamples
                # sets feasibility flag
                index = 0
                for i in range(IListCpp.Count.GetValue(points)):
                    pt = IListCpp.__getitem__(points,index)
                    index = index + 1
                    pt.Feasible = True
                    for c in self._constraints:
                        if c.Compute(pt) > 0:
                            pt.Feasible = False
                            numFeasibles = numFeasibles - 1
                            break
                if numFeasibles < nbsets:
                    maxfronts = numFeasibles
                else:
                    maxfronts = nbsets
                if numFeasibles != numberOfSamples:
                    maxfronts = maxfronts - 1
                cpt = 1
                index = 0
                for i in range(IListCpp.Count.GetValue(points)):
                    pt = IListCpp.__getitem__(points,index)
                    index = index + 1
                    if pt.Feasible == True:
                        if cpt <= maxfronts:
                             pt.ParetoFrontIndex = cpt # at least one sample per paretofront
                             cpt = cpt + 1
                        else:
                             pt.ParetoFrontIndex = random.randint(1,maxfronts)
                    else:
                        pt.ParetoFrontIndex = maxfronts + 1
    # generates random point 
    def GenerateRandomPoint(self,ptIdx):
        point = PointHelper(DXOptimizationPoint())
        point.ID = 'DP %(id)04d' % {"id" : ptIdx}
        for p in self._parameters:
            if p._type == 2:
                if p._customproperties is None or p._customproperties["Distribution"] is None:
                    rand = random.random()
                    val = p._lowerbound + rand * (p._upperbound - p._lowerbound)
                elif p._customproperties["Distribution"] == "Uniform":
                    val = random.uniform(p._lowerbound,p._upperbound)
                elif p._customproperties["Distribution"] == "Triangular" and p._customproperties["PeakLocation"] != "":
                    val = random.triangular(p._lowerbound,p._upperbound,p._customproperties["PeakLocation"])
                elif p._customproperties["Distribution"] == "Normal" and p._customproperties["Mean"] != "" and p._customproperties["StandardDeviation"] != "":
                    val = random.normalvariate(p._customproperties["Mean"],p._customproperties["StandardDeviation"])
                else:
                    rand = random.random()
                    val = p._lowerbound + rand * (p._upperbound - p._lowerbound)
                point.put_Value(p._id,val)
        return point       
      
    # computes the cost function of the optimization point 
    # = sum of cost function of objectives and constraints  
    def CostFunction(self,pt):
         cost1 = 0.0
         for o in self._objectives: 
           cost1 = cost1 + o.Compute(pt)
         cost2 = 0.0            
         for c in self._constraints:
           cost2 = cost2 + c.Compute(pt)
         return cost1 + cost2
         
# end of the class

